local lunajson = require('lunajson') -- luarocks install lunajson
local subproc = require('subproc') -- luarocks install subproc
local posix = {
	sys = {
		socket = require('posix.sys.socket') -- luarocks install luaposix
	}
}
local pss = posix.sys.socket

local Conn = {}
Conn.__index = Conn

local MSGTYPE = {
	RUN_COMMAND = 0,
	GET_WORKSPACES = 1,
	SUBSCRIBE = 2,
	GET_OUTPUTS = 3,
	GET_TREE = 4,
	GET_MARKS = 5,
	GET_BAR_CONFIG = 6,
	GET_VERSION = 7,
	GET_BINDING_MODES = 8,
	GET_CONFIG = 9,
	SEND_TICK = 10,
	SYNC = 11,
	GET_BINDING_STATE = 12
}

local I3_MAGIC = 'i3-ipc'

--[[ given message-type t and messag m, pack it into i3's binary format ]]
local function pack_msg(t, msg)
	--[[
		format of an i3 message:
		'i3-ipc'
		32-bit length of message payload (native order)
		32-bit type (native order)
	]]
	return I3_MAGIC .. string.pack('I4I4', #msg, t) .. msg
end

--[[ find the path for the unix socket ]]
local function find_socket_path()
	local sock = os.getenv('I3SOCK') or os.getenv('SWAYSOCK')
	if sock ~= nil then
		return sock
	end

	for _, bin in ipairs({'i3', 'sway'}) do
		out, _, status = subproc(bin, '--get-socketpath')
		if status == 0 then
			sock = out
			break
		end
	end
	return sock
end

--[[
	Connect to the i3 or sway socket, using the provided socket or autodetecting
	if none provided
]]
local function connect(socket_path, auto_recon)
	if socket_path == nil then
		socket_path = find_socket_path()
	end

	if not socket_path then
		error('error: no i3/sway socket provided and none could be detected')
	end

	local socket = pss.socket(pss.AF_UNIX, pss.SOCK_STREAM, 0)
	if type(socket) ~= 'number' then
		error('cannot create socket: ' .. socket)
	end

	-- 1 second timeout
	pss.setsockopt(socket, pss.SOL_SOCKET, pss.SO_SNDTIMEO, 1, 0)

	local socket_addr = {
		family = pss.AF_UNIX,
		path = socket_path
	}

	local err = pss.connect(socket, socket_addr)
	if err ~= 0 then
		error('cannot connect to socket: ' .. err)
	end

	local i3conn = {
		socket_addr = sock_addr,
		socket = socket
	}
	setmetatable(i3conn, Conn)
	return i3conn
end

--[[
	send a message to the connection. then receive a response
]]
function Conn:send_msg(t, msg)
	local packed = pack_msg(t, msg)
	local bytes_written = pss.send(self.socket, packed)
	if bytes_written ~= #packed then
		print('err: expected to write ' .. #packed .. ' bytes but wrote ' .. bytes_written)
	end

	-- read the magic and two lengths
	local resp_header = pss.recv(self.socket, #I3_MAGIC + 8)

	local resp_len, resp_type = string.unpack('I4I4', resp_header, #I3_MAGIC + 1)
	local resp_pld
	if resp_len == 0 then
		resp_pld = nil
	else
		resp_pld = pss.recv(self.socket, resp_len)
	end
	return resp_pld, resp_type
end

function Conn:command(payload)
	local resp = self:send_msg(MSGTYPE.RUN_COMMAND, payload)
	if resp then
		return lunajson.decode(resp)
	else
		return {}
	end
end

function Conn:get_tree()
	local resp = self:send_msg(MSGTYPE.GET_TREE, '')
	local tree = lunajson.decode(resp)

	local function add_parent_fields(t)
		for _,n in ipairs(t.nodes) do
			n.parent = t
			add_parent_fields(n)
		end
	end
	
	add_parent_fields(tree)
	return tree

	-- add parent markers
end



local M = {
	connect = connect,
	Connection = Conn
}

return M
