-- this is a port of i3-window-tool (python tool) because python takes way too
-- long to start up.

-- It's basically to make the directional window-switching commands work more
-- intuitively on i3/sway.
local i3connection = require('i3connection')

local conn = i3connection.connect()
local tree = conn:get_tree()

--[[ Iterate through the children of "node" in most recently focused order. ]]
local function focused_order(node)
	return coroutine.wrap(function()
		for _,focus_id in ipairs(node.focus) do
			for _,n in ipairs(node.nodes) do
				if n.id == focus_id then
					coroutine.yield(n)
				end
			end
		end
	end)
end

--[[ Return the most recently focused child of "node". ]]
local function focused_child(node)
	return focused_order(node)()
end

local function cur_focus(node)
	if node.focused then
		return node
	end
	for _,n in ipairs(node.nodes) do
		local focus = cur_focus(n)
		if focus ~= nil then
			return focus
		end
	end
	return nil
end

--[[
	Return true if container "new" can reasonably be considered to be in
	direction "direction" of container "old".
]]
local function is_in_line(old, new, dir)
	if dir == 'up' or dir == 'down' then
		return new.rect.x <= old.rect.x + old.rect.width * 0.9 and
			new.rect.x + new.rect.width >= old.rect.x + old.rect.width * 0.1
	elseif dir == 'left' or dir == 'right' then
		return new.rect.y <= old.rect.y + old.rect.height * 0.9 and
			new.rect.y + new.rect.height >= old.rect.y + old.rect.height * 0.1
	end
end

--[[
	Return the output in direction "direction" of window "window" on output
	"output".
]]
local function output_in_direction(out, win, dir)
	for new in focused_order(tree) do
		if new.name == "__i3" or is_in_line(win, new, dir) then
			goto continue
		end

		local orct = output.rect
		local nrct = new.rect
		if (dir == 'left' and nrct.x + nrct.width == orct.x) or
				(dir == 'right' and nrct.x == orct.x + orct.width) or
				(dir == 'up' and nrct.y + nrct.height == orct.y) or
				(dir == 'down' and nrct.y == orct.y + orct.height) then
			return new
		end
		::continue::
	end

	return nil
end

--[[
	Return the leaf that is *visually* in direction "direction" of the node
	"window". If there are multiple candidates, (usually) pick the most recently
	focused one.
]]
local function leaf_in_direction(win, dir)
	local splitx, last, first, delta
	
	if dir == 'left' or dir == 'right' then
		splitx = 'splith'
	elseif dir == 'up' or dir == 'down' then
		splitx = 'splitv' 
	end

	if dir == 'down' or dir == 'right' then
		last = function(xs)
			return xs[#xs]
		end
		first = function(xs)
			return xs[1]
		end
		delta = 1
	elseif dir == 'up' or dir == 'left' then
		last = function(xs)
			return xs[1]
		end
		first = function(xs)
			return xs[#xs]
		end
		delta = -1
	else
		usage()
	end

	local node = win

	-- Find innermost parent in which we can move in the desired direction,
	-- then move.
	while true do
		local par = node.parent
		if node.type == 'workspace' then
			node = output_in_direction(par.parent, win, dir)
			if not node then
				return nil
			end
			for _,n in ipairs(node.nodes) do
				if n.type ~= 'dockarea' then
					node = n
					break
				end
			end
			node = focused_child(node)
			goto node_found
		elseif node.type == 'floating_con' then
			return nil
		elseif node.type ~= 'con' then
			error('not con')
		elseif not par then
			error('no parent')
		elseif par.layout == splitx and last(par.nodes) ~= node then
			for idx,n in ipairs(par.nodes) do
				if n == node then
					node = par.nodes[idx + delta]
					goto node_found
				end
			end
		else
			node = par
		end
	end
	::node_found::

	-- Find an appropriate leaf to focus
	while #node.nodes ~= 0 do
		if node.layout == 'tabbed' or node.layout == 'stacked' then
			-- Choose most recently focused tab.
			node = focused_child(node)
		elseif node.layout == splitx then
			-- Choose nearest child to original container.
			node = first(node.nodes)
		else
			-- Choose the most recently focused child that can reasonably be
			-- considered adjacent to the original container.
			for new in focused_order(node) do
				if is_in_line(win, new, dir) then
					node = new
					goto recent_focus_found
				end
			end

			-- No appropriate child found, just focus any one.
			node = node.nodes[1]

			::recent_focus_found::
		end
	end

	return node
end

--[[
	Focus the leaf that is *visually* in direction "direction". If there are
	multiple candidates, (usually) pick the most recently focused one.
]]
local function focus_direction(dir)
	local focused = cur_focus(tree)
	if focused == nil then
		error('no current focus?')
	end
	local node = leaf_in_direction(focused, dir)
	if node then
		conn:command(string.format('[con_id="%d"] focus', node.id))
	end
end

--[[
	Move the current container in the given direction. I the direct parent is a
	tabbed or stacked container, escape from it.
]]
local function move_direction(dir, amount)
	local focused = cur_focus(tree)

	local tabx, taby
	if dir == 'left' or dir == 'right' then
		tabx = 'tabbed'
		taby = 'stacked'
	elseif dir == 'up' or dir == 'down' then
		tabx = 'stacked'
		taby = 'tabbed'
	else
		usage()
	end

	local parent = focused.parent
	if parent and parent.layout == tabx then
		for _, n in ipairs(parent.nodes) do
			if n ~= focused then
				local sib = n
				local cmd = string.format(
					'layout %s; move %s %d px; [con_id="%d"] layout %s',
					taby, dir, amount, sib.id, tabx
				)
				conn:command(cmd)
				return
			end
		end
	end

	conn:command(string.format('move %s %d px', dir, amount))
end
			
--[[
	Swap with the leaf that is *visually* in direction "direction". If there are
	multiple candidates, (usually) pick the most recently focused one.
]]
local function swap_direction(dir)
	local focused = cur_focus(tree)
	local node = leaf_in_direction(focused, dir)
	if node then
		conn:command(string.format('swap container with con_id %d', node.id))
	end
end		

--[[
	Cycle through the innermost stacked or tabbed ancestor container, or through
	floating containers.
]]
local function focus_tab(dir)
	local delta
	if dir == 'next' then
		delta = 1
	elseif dir == 'prev' then
		delta = -1
	else
		usage()
	end

	local node = cur_focus(tree)

	-- Find innermost tabbed or stacked container, or detect floating.
	while true do
		local p = node.parent
		if not p or node.type ~= 'con' then
			return
		elseif p.layout == 'tabbed' or
			p.layout 'stacked' or
			p.type == 'floating_con' then
			break
		else
			node = p
		end
	end

	local parent_nodes
	if p.type == 'floating_con' then
		node = p
		p = node.parent

		-- The order of floating_nodes is not useful, sort it somehow.
		parent_nodes = p.floating_nodes
	else
		parent_nodes = p.nodes
	end

	for idx, n in ipairs(parent_nodes) do
		if n == node then
			node = parent_nodes[(idx + delta) % #parent_nodes]
		end
	end

	-- Find most recently focused leaf in new tab.
	while node.nodes do
		node = focused_child(node)
	end

	conn:command(string.format('[con_id="%d"] focus', node.id))
end

--[[
	Move the innermost stacked or tabbed ancestor container.
]]
local function move_tab(dir)
	local delta
	if dir == 'next' then
		delta = 1
	elseif dir == 'prev' then
		delta = -1
	else
		usage()
	end

	local p
	local node = cur_focus(tree)
	while true do
		p = node.parent
		if not p or node.type ~= 'con' then
			return
		elseif p.layout == 'tabbed' or p.layout == 'stacked' then
			break
		else
			node = p
		end
	end

	for n,idx in ipairs(p.nodes) do
		if n == node then
			if 1 <= idx + delta <= #p.nodes then
				local other = p.nodes[idx + delta]
				conn:command(string.format('[con_id="%d"] swap contained with con_id %d', node.id, other.id))
			end
		end
	end
end

local function usage()
	print("Usage: i3win [focus|swap] [up|down|left|right]")
	print("       i3win move [up|down|left|right] AMOUNT")
	print("       i3win [tab-focus|tab-move] [prev|next]")
	os.exit(1)
end

local function main(cmd, dir)
	if cmd == 'focus' then
		focus_direction(dir)
	elseif cmd == 'swap' then
		swap_direction(dir)
	elseif cmd == 'move' then
		move_direction(dir, arg)
	elseif cmd == 'tab-focus' then
		focus_tab(dir)
	elseif cmd == 'tab-move' then
		move_tab(dir)
	else
		usage()
	end
end

local argv = table.pack(...)
if #argv ~= 2 then
	usage()
end
main(table.unpack(argv))

